@@ -36,6 +36,8 @@
 #  ifndef _CRT_NONSTDC_NO_DEPRECATE
 #    define _CRT_NONSTDC_NO_DEPRECATE
 #  endif
+/* enable predefined math constants M_* for MS Visual Studio workaround */
+#  define _USE_MATH_DEFINES
 #endif
 
 /* standard inclusions */
@@ -48,21 +50,21 @@
 #define C_NAMESPACE extern "C"
 #define C_NAMESPACE_VAR extern "C"
 extern "C" {
-#else    
+#else
 #define C_NAMESPACE extern
 #define C_NAMESPACE_VAR
 #endif
 
 #ifndef NULL
-#  define NULL	0
+#  define NULL 0
 #endif
 
 #ifndef FALSE
-#  define FALSE	0
+#  define FALSE 0
 #endif
 
 #ifndef TRUE
-#  define TRUE	1
+#  define TRUE  1
 #endif
 
 #ifndef MAX
@@ -74,11 +76,12 @@ extern "C" {
 #  define ABS(x)        ((x<0) ? (-1*(x)) : x)
 #endif
 
-    /* maximum path/filename */
+/* maximum path/filename */
 #ifndef MAX_PATH_FILENAME
 #define MAX_PATH_FILENAME 1024
 #endif
-	/* prototype hypot for systems where absent */
+
+/* prototype hypot for systems where absent */
 #ifndef _WIN32
 extern double hypot(double, double);
 #endif
@@ -92,11 +95,24 @@ extern double hypot(double, double);
 #  define hypot _hypot
 #endif
 
-	/* some useful constants */
-#define HALFPI		1.5707963267948966
-#define FORTPI		0.78539816339744833
-#define PI		3.14159265358979323846
-#define TWOPI		6.2831853071795864769
+/* If we still haven't got M_PI*, we rely on our own defines.
+ * For example, this is necessary when compiling with gcc and
+ * the -ansi flag.
+ */
+#ifndef M_PI
+#define M_PI            3.14159265358979310
+#define M_PI_2          1.57079632679489660
+#define M_PI_4          0.78539816339744828
+#endif
+
+/* some more useful math constants and aliases */
+#define M_FORTPI        M_PI_4                   /* pi/4 */
+#define M_HALFPI        M_PI_2                   /* pi/2 */
+#define M_PI_HALFPI     4.71238898038468985769   /* 1.5*pi */
+#define M_TWOPI         6.28318530717958647693   /* 2*pi */
+#define M_TWO_D_PI      M_2_PI                   /* 2/pi */
+#define M_TWOPI_HALFPI  7.85398163397448309616   /* 2.5*pi */
+
 
 /* maximum tag id length for +init and default files */
 #ifndef ID_TAG_MAX
@@ -123,7 +139,7 @@ struct projFileAPI_t;
 
 /* proj thread context */
 typedef struct {
-    int	    last_errno;
+    int     last_errno;
     int     debug_level;
     void    (*logger)(void *, int, const char *);
     void    *app_data;
@@ -132,8 +148,8 @@ typedef struct {
 
 /* datum_type values */
 #define PJD_UNKNOWN   0
-#define PJD_3PARAM    1   
-#define PJD_7PARAM    2   
+#define PJD_3PARAM    1
+#define PJD_7PARAM    2
 #define PJD_GRIDSHIFT 3
 #define PJD_WGS84     4   /* WGS84 (or anything considered equivelent) */
 
@@ -143,10 +159,10 @@ typedef struct {
 #define PJD_ERR_GRID_AREA           -48
 #define PJD_ERR_CATALOG             -49
 
-#define USE_PROJUV 
+#define USE_PROJUV
 
 typedef struct { double u, v; } projUV;
-typedef struct { double r, i; }	COMPLEX;
+typedef struct { double r, i; } COMPLEX;
 typedef struct { double u, v, w; } projUVW;
 
 #ifndef PJ_LIB__
@@ -163,125 +179,143 @@ typedef struct { double lam, phi, z; } LPZ;
 
 typedef union { double  f; int  i; char *s; } PROJVALUE;
 struct PJconsts;
-    
+
 struct PJ_LIST {
-	char	*id;		/* projection keyword */
-	struct PJconsts	*(*proj)(struct PJconsts*);/* projection entry point */
-	char 	* const *descr;	/* description text */
+    char             *id;                         /* projection keyword */
+    struct PJconsts  *(*proj)(struct PJconsts*);  /* projection entry point */
+    char * const     *descr;                      /* description text */
 };
+
+/* Merging this into the PJ_LIST infrastructure is tempting, but may imply ABI breakage. Perhaps at next major version? */
+struct PJ_SELFTEST_LIST {
+    char    *id;                    /* projection keyword */
+    int     (* testfunc)(void);     /* projection entry point */
+};
+
 struct PJ_ELLPS {
-	char	*id;	/* ellipse keyword name */
-	char	*major;	/* a= value */
-	char	*ell;	/* elliptical parameter */
-	char	*name;	/* comments */
+    char    *id;        /* ellipse keyword name */
+    char    *major;     /* a= value */
+    char    *ell;       /* elliptical parameter */
+    char    *name;      /* comments */
 };
 struct PJ_UNITS {
-	char	*id;	/* units keyword */
-	char	*to_meter;	/* multiply by value to get meters */
-	char	*name;	/* comments */
+    char    *id;        /* units keyword */
+    char    *to_meter;  /* multiply by value to get meters */
+    char    *name;      /* comments */
 };
 
 struct PJ_DATUMS {
-    char    *id;     /* datum keyword */
-    char    *defn;   /* ie. "to_wgs84=..." */
-    char    *ellipse_id; /* ie from ellipse table */
-    char    *comments; /* EPSG code, etc */
+    char    *id;        /* datum keyword */
+    char    *defn;      /* ie. "to_wgs84=..." */
+    char    *ellipse_id;/* ie from ellipse table */
+    char    *comments;  /* EPSG code, etc */
 };
 
 struct PJ_PRIME_MERIDIANS {
-    char    *id;     /* prime meridian keyword */
-    char    *defn;   /* offset from greenwich in DMS format. */
+    char    *id;        /* prime meridian keyword */
+    char    *defn;      /* offset from greenwich in DMS format. */
 };
 
 typedef struct {
     double ll_long;      /* lower left corner coordinates (radians) */
     double ll_lat;
     double ur_long;      /* upper right corner coordinates (radians) */
-    double ur_lat; 
+    double ur_lat;
 } PJ_Region;
 
 struct DERIVS {
-    double x_l, x_p; /* derivatives of x for lambda-phi */
-    double y_l, y_p; /* derivatives of y for lambda-phi */
+    double x_l, x_p;    /* derivatives of x for lambda-phi */
+    double y_l, y_p;    /* derivatives of y for lambda-phi */
 };
-    
-struct FACTORS {
-	struct DERIVS der;
-	double h, k;	/* meridinal, parallel scales */
-	double omega, thetap;	/* angular distortion, theta prime */
-	double conv;	/* convergence */
-	double s;		/* areal scale factor */
-	double a, b;	/* max-min scale error */
-	int code;		/* info as to analytics, see following */
-};
-#define IS_ANAL_XL_YL 01	/* derivatives of lon analytic */
-#define IS_ANAL_XP_YP 02	/* derivatives of lat analytic */
-#define IS_ANAL_HK	04		/* h and k analytic */
-#define IS_ANAL_CONV 010	/* convergence analytic */
-    /* parameter list struct */
-typedef struct ARG_list {
-	struct ARG_list *next;
-	char used;
-	char param[1]; } paralist;
-	/* base projection data structure */
 
+struct FACTORS {
+    struct DERIVS der;
+    double h, k;        /* meridinal, parallel scales */
+    double omega, thetap;   /* angular distortion, theta prime */
+    double conv;        /* convergence */
+    double s;           /* areal scale factor */
+    double a, b;        /* max-min scale error */
+    int code;           /* info as to analytics, see following */
+};
+
+#define IS_ANAL_XL_YL 01    /* derivatives of lon analytic */
+#define IS_ANAL_XP_YP 02    /* derivatives of lat analytic */
+#define IS_ANAL_HK    04    /* h and k analytic */
+#define IS_ANAL_CONV 010    /* convergence analytic */
+
+/* parameter list struct */
+typedef struct ARG_list {
+    struct ARG_list *next;
+    char used;
+    char param[1]; } paralist;
+
+/* base projection data structure */
+#ifdef PJ_LIB__
+    /* we need this forward declaration in order to be able to add a
+       pointer to struct opaque to the typedef struct PJconsts below */
+    struct pj_opaque;
+#endif
 
 typedef struct PJconsts {
     projCtx_t *ctx;
-	XY  (*fwd)(LP, struct PJconsts *);
-	LP  (*inv)(XY, struct PJconsts *);
-        XYZ (*fwd3d)(LPZ, struct PJconsts *);
-        LPZ (*inv3d)(XYZ, struct PJconsts *); 
-	void (*spc)(LP, struct PJconsts *, struct FACTORS *);
-	void (*pfree)(struct PJconsts *);
-	const char *descr;
-	paralist *params;   /* parameter list */
-	int over;   /* over-range flag */
-	int geoc;   /* geocentric latitude flag */
-        int is_latlong; /* proj=latlong ... not really a projection at all */
-        int is_geocent; /* proj=geocent ... not really a projection at all */
-	double
-		a,  /* major axis or radius if es==0 */
-                a_orig, /* major axis before any +proj related adjustment */
-		es, /* e ^ 2 */
-                es_orig, /* es before any +proj related adjustment */
-		e,  /* eccentricity */
-		ra, /* 1/A */
-		one_es, /* 1 - e^2 */
-		rone_es, /* 1/one_es */
-		lam0, phi0, /* central longitude, latitude */
-		x0, y0, /* easting and northing */
-		k0,	/* general scaling factor */
-		to_meter, fr_meter; /* cartesian scaling */
-    
-        int     datum_type; /* PJD_UNKNOWN/3PARAM/7PARAM/GRIDSHIFT/WGS84 */
-        double  datum_params[7];
-        struct _pj_gi **gridlist;
-        int     gridlist_count;
+    XY  (*fwd)(LP, struct PJconsts *);
+    LP  (*inv)(XY, struct PJconsts *);
+    XYZ (*fwd3d)(LPZ, struct PJconsts *);
+    LPZ (*inv3d)(XYZ, struct PJconsts *);
+    void (*spc)(LP, struct PJconsts *, struct FACTORS *);
+    void (*pfree)(struct PJconsts *);
 
-        int     has_geoid_vgrids;
-        struct _pj_gi **vgridlist_geoid;
-        int     vgridlist_geoid_count;
-        double  vto_meter, vfr_meter;
+    const char *descr;
+    paralist *params;           /* parameter list */
+    int over;                   /* over-range flag */
+    int geoc;                   /* geocentric latitude flag */
+    int is_latlong;             /* proj=latlong ... not really a projection at all */
+    int is_geocent;             /* proj=geocent ... not really a projection at all */
+    double a;                   /* major axis or radius if es==0 */
+    double a_orig;              /* major axis before any +proj related adjustment */
+    double es;                  /* e ^ 2 */
+    double es_orig;             /* es before any +proj related adjustment */
+    double e;                   /* eccentricity */
+    double ra;                  /* 1/A */
+    double one_es;              /* 1 - e^2 */
+    double rone_es;             /* 1/one_es */
+    double lam0, phi0;          /* central longitude, latitude */
+    double x0, y0;              /* easting and northing */
+    double k0;                  /* general scaling factor */
+    double to_meter, fr_meter;  /* cartesian scaling */
 
-        double  from_greenwich; /* prime meridian offset (in radians) */
-        double  long_wrap_center; /* 0.0 for -180 to 180, actually in radians*/
-        int     is_long_wrap_set;
-        char    axis[4];
+    int     datum_type;         /* PJD_UNKNOWN/3PARAM/7PARAM/GRIDSHIFT/WGS84 */
+    double  datum_params[7];
+    struct _pj_gi **gridlist;
+    int     gridlist_count;
 
-        /* New Datum Shift Grid Catalogs */
-        char   *catalog_name;
-        struct _PJ_GridCatalog *catalog;
-    
-        double   datum_date;
-    
-        struct _pj_gi *last_before_grid;
-        PJ_Region     last_before_region;
-        double        last_before_date;
+    int     has_geoid_vgrids;
+    struct _pj_gi **vgridlist_geoid;
+    int     vgridlist_geoid_count;
+    double  vto_meter, vfr_meter;
 
-        struct _pj_gi *last_after_grid;
-        PJ_Region     last_after_region;
-        double        last_after_date;
+    double  from_greenwich;     /* prime meridian offset (in radians) */
+    double  long_wrap_center;   /* 0.0 for -180 to 180, actually in radians*/
+    int     is_long_wrap_set;
+    char    axis[4];
+
+    /* New Datum Shift Grid Catalogs */
+    char   *catalog_name;
+    struct _PJ_GridCatalog *catalog;
+
+    double   datum_date;
+
+    struct _pj_gi *last_before_grid;
+    PJ_Region     last_before_region;
+    double        last_before_date;
+
+    struct _pj_gi *last_after_grid;
+    PJ_Region     last_after_region;
+    double        last_after_date;
+
+#ifdef PJ_LIB__
+        struct pj_opaque *opaque;
+#endif
 
 #ifdef PROJ_PARMS__
 PROJ_PARMS__
@@ -291,24 +325,16 @@ PROJ_PARMS__
 /* public API */
 #include "proj_api.h"
 
+
 /* Generate pj_list external or make list from include file */
+
 #ifndef USE_PJ_LIST_H
 extern struct PJ_LIST pj_list[];
-#else
-#define PROJ_HEAD(id, name) \
-    struct PJconsts *pj_##id(struct PJconsts*); extern char * const pj_s_##id;
-    
-#include "pj_list.h"
-#undef PROJ_HEAD
-#define PROJ_HEAD(id, name) {#id, pj_##id, &pj_s_##id},
-	struct PJ_LIST
-pj_list[] = {
-#include "pj_list.h"
-		{0,     0,  0},
-	};
-#undef PROJ_HEAD
+extern struct PJ_SELFTEST_LIST pj_selftest_list[];
 #endif
 
+
+
 #ifndef PJ_ELLPS__
 extern struct PJ_ELLPS pj_ellps[];
 #endif
@@ -323,16 +349,16 @@ extern struct PJ_PRIME_MERIDIANS pj_prime_meridians[];
 #endif
 
 #ifdef PJ_LIB__
-    /* repetitive projection code */
+/* repetitive projection code */
 #define PROJ_HEAD(id, name) static const char des_##id [] = name
 #define ENTRYA(name) \
-        C_NAMESPACE_VAR const char * const pj_s_##name = des_##name; \
-	C_NAMESPACE PJ *pj_##name(PJ *P) { if (!P) { \
-	if( (P = (PJ*) pj_malloc(sizeof(PJ))) != NULL) { \
+    C_NAMESPACE_VAR const char * const pj_s_##name = des_##name; \
+    C_NAMESPACE PJ *pj_##name(PJ *P) { if (!P) { \
+    if( (P = (PJ*) pj_malloc(sizeof(PJ))) != NULL) { \
         memset( P, 0, sizeof(PJ) ); \
-	P->pfree = freeup; P->fwd = 0; P->inv = 0; \
+    P->pfree = freeup; P->fwd = 0; P->inv = 0; \
         P->fwd3d = 0; P->inv3d = 0; \
-	P->spc = 0; P->descr = des_##name;
+    P->spc = 0; P->descr = des_##name;
 #define ENTRYX } return P; } else {
 #define ENTRY0(name) ENTRYA(name) ENTRYX
 #define ENTRY1(name, a) ENTRYA(name) P->a = 0; ENTRYX
@@ -350,27 +376,66 @@ extern struct PJ_PRIME_MERIDIANS pj_prime_meridians[];
 #define INVERSE3D(name) static LPZ name(XYZ xyz, PJ *P) {LPZ lpz = {0.0, 0.0, 0.0}
 #define FREEUP static void freeup(PJ *P) {
 #define SPECIAL(name) static void name(LP lp, PJ *P, struct FACTORS *fac)
+#define ELLIPSOIDAL(P) ((P->es==0)? (FALSE): (TRUE))
+
+/* cleaned up alternative to most of the "repetitive projection code" macros */
+#define PROJECTION(name)                                     \
+pj_projection_specific_setup_##name (PJ *P);                 \
+C_NAMESPACE_VAR const char * const pj_s_##name = des_##name; \
+C_NAMESPACE PJ *pj_##name (PJ *P) {                          \
+    if (P)                                                   \
+        return pj_projection_specific_setup_##name (P);      \
+    P = (PJ*) pj_calloc (1, sizeof(PJ));                     \
+    if (0==P)                                                \
+        return 0;                                            \
+    P->pfree = freeup;                                       \
+    P->descr = des_##name;                                   \
+    return P;                                                \
+}                                                            \
+PJ *pj_projection_specific_setup_##name (PJ *P)
+
 #endif
+
+
+int pj_generic_selftest (
+    char *e_args,
+    char *s_args,
+    double tolerance_xy,
+    double tolerance_lp,
+    int n_fwd,
+    int n_inv,
+    LP *fwd_in,
+    XY *e_fwd_expect,
+    XY *s_fwd_expect,
+    XY *inv_in,
+    LP *e_inv_expect,
+    LP *s_inv_expect
+);
+
+
+
+
 #define MAX_TAB_ID 80
 typedef struct { float lam, phi; } FLP;
 typedef struct { int lam, phi; } ILP;
 
 struct CTABLE {
-	char id[MAX_TAB_ID]; /* ascii info */
-	LP ll;      /* lower left corner coordinates */
-	LP del;     /* size of cells */
-	ILP lim;    /* limits of conversion matrix */
-	FLP *cvs;   /* conversion matrix */
+    char id[MAX_TAB_ID];    /* ascii info */
+    LP ll;                  /* lower left corner coordinates */
+    LP del;                 /* size of cells */
+    ILP lim;                /* limits of conversion matrix */
+    FLP *cvs;               /* conversion matrix */
 };
 
 typedef struct _pj_gi {
-    char *gridname;   /* identifying name of grid, eg "conus" or ntv2_0.gsb */
-    char *filename;   /* full path to filename */
-    
-    const char *format; /* format of this grid, ie "ctable", "ntv1", 
+    char *gridname;     /* identifying name of grid, eg "conus" or ntv2_0.gsb */
+    char *filename;     /* full path to filename */
+
+    const char *format; /* format of this grid, ie "ctable", "ntv1",
                            "ntv2" or "missing". */
 
-    int   grid_offset; /* offset in file, for delayed loading */
+    int   grid_offset;  /* offset in file, for delayed loading */
+    int   must_swap;    /* only for NTv2 */
 
     struct CTABLE *ct;
 
@@ -380,18 +445,18 @@ typedef struct _pj_gi {
 
 typedef struct {
     PJ_Region region;
-    int  priority; /* higher used before lower */
-    double date; /* year.fraction */
-    char *definition; /* usually the gridname */
+    int  priority;      /* higher used before lower */
+    double date;        /* year.fraction */
+    char *definition;   /* usually the gridname */
 
     PJ_GRIDINFO  *gridinfo;
-    int available; /* 0=unknown, 1=true, -1=false */
+    int available;      /* 0=unknown, 1=true, -1=false */
 } PJ_GridCatalogEntry;
 
 typedef struct _PJ_GridCatalog {
     char *catalog_name;
 
-    PJ_Region region; /* maximum extent of catalog data */
+    PJ_Region region;   /* maximum extent of catalog data */
 
     int entry_count;
     PJ_GridCatalogEntry *entries;
@@ -413,6 +478,7 @@ int pj_ell_set(projCtx ctx, paralist *, double *, double *);
 int pj_datum_set(projCtx,paralist *, PJ *);
 int pj_prime_meridian_set(paralist *, PJ *);
 int pj_angular_units_set(paralist *, PJ *);
+void pj_prepare (PJ *P, const char *description, void (*freeup)(struct PJconsts *), size_t sizeof_struct_opaque);
 
 paralist *pj_clone_paralist( const paralist* );
 paralist*pj_search_initcache( const char *filekey );
@@ -434,18 +500,18 @@ COMPLEX pj_zpolyd1(COMPLEX, COMPLEX *, int, COMPLEX *);
 int pj_deriv(LP, double, PJ *, struct DERIVS *);
 int pj_factors(LP, PJ *, double, struct FACTORS *);
 
-struct PW_COEF {/* row coefficient structure */
-    int m;		/* number of c coefficients (=0 for none) */
-    double *c;	/* power coefficients */
+struct PW_COEF {    /* row coefficient structure */
+    int m;          /* number of c coefficients (=0 for none) */
+    double *c;      /* power coefficients */
 };
- 
+
 /* Approximation structures and procedures */
-typedef struct {	/* Chebyshev or Power series structure */
-	projUV a, b;		/* power series range for evaluation */
-					/* or Chebyshev argument shift/scaling */
-	struct PW_COEF *cu, *cv;
-	int mu, mv;		/* maximum cu and cv index (+1 for count) */
-	int power;		/* != 0 if power series, else Chebyshev */
+typedef struct {    /* Chebyshev or Power series structure */
+    projUV a, b;    /* power series range for evaluation */
+                    /* or Chebyshev argument shift/scaling */
+    struct PW_COEF *cu, *cv;
+    int mu, mv;     /* maximum cu and cv index (+1 for count) */
+    int power;      /* != 0 if power series, else Chebyshev */
 } Tseries;
 Tseries *mk_cheby(projUV, projUV, double, projUV *, projUV (*)(projUV), int, int, int);
 projUV bpseval(projUV, Tseries *);
@@ -456,6 +522,7 @@ void **vector2(int, int, int);
 void freev2(void **v, int nrows);
 int bchgen(projUV, projUV, int, int, projUV **, projUV(*)(projUV));
 int bch2bps(projUV, projUV, projUV **, int, int);
+
 /* nadcon related protos */
 LP nad_intr(LP, struct CTABLE *);
 LP nad_cvt(LP, int, struct CTABLE *);
@@ -469,15 +536,15 @@ void nad_free(struct CTABLE *);
 /* higher level handling of datum grid shift files */
 
 int pj_apply_vgridshift( PJ *defn, const char *listname,
-                         PJ_GRIDINFO ***gridlist_p, 
+                         PJ_GRIDINFO ***gridlist_p,
                          int *gridlist_count_p,
-                         int inverse, 
+                         int inverse,
                          long point_count, int point_offset,
                          double *x, double *y, double *z );
-int pj_apply_gridshift_2( PJ *defn, int inverse, 
+int pj_apply_gridshift_2( PJ *defn, int inverse,
                           long point_count, int point_offset,
                           double *x, double *y, double *z );
-int pj_apply_gridshift_3( projCtx ctx, 
+int pj_apply_gridshift_3( projCtx ctx,
                           PJ_GRIDINFO **gridlist, int gridlist_count,
                           int inverse, long point_count, int point_offset,
                           double *x, double *y, double *z );
@@ -492,15 +559,15 @@ void pj_gridinfo_free( projCtx, PJ_GRIDINFO * );
 PJ_GridCatalog *pj_gc_findcatalog( projCtx, const char * );
 PJ_GridCatalog *pj_gc_readcatalog( projCtx, const char * );
 void pj_gc_unloadall( projCtx );
-int pj_gc_apply_gridshift( PJ *defn, int inverse, 
+int pj_gc_apply_gridshift( PJ *defn, int inverse,
                            long point_count, int point_offset,
                            double *x, double *y, double *z );
-int pj_gc_apply_gridshift( PJ *defn, int inverse, 
+int pj_gc_apply_gridshift( PJ *defn, int inverse,
                            long point_count, int point_offset,
                            double *x, double *y, double *z );
 
-PJ_GRIDINFO *pj_gc_findgrid( projCtx ctx, 
-                             PJ_GridCatalog *catalog, int after, 
+PJ_GRIDINFO *pj_gc_findgrid( projCtx ctx,
+                             PJ_GridCatalog *catalog, int after,
                              LP location, double date,
                              PJ_Region *optional_region,
                              double *grid_date );
@@ -520,6 +587,7 @@ struct PJ_ELLPS *pj_get_ellps_ref( void );
 struct PJ_DATUMS *pj_get_datums_ref( void );
 struct PJ_UNITS *pj_get_units_ref( void );
 struct PJ_LIST  *pj_get_list_ref( void );
+struct PJ_SELFTEST_LIST  *pj_get_selftest_list_ref ( void );
 struct PJ_PRIME_MERIDIANS  *pj_get_prime_meridians_ref( void );
 
 double pj_atof( const char* nptr );
